(*----------------------------------------------------------------------------*)
(* National Language Support                                                  *)
(* Start of this file: 7.1.2002                                               *)
(* Author: Joachim Pimiskern, 2002-2004                                       *)
(* Status: Freeware. Use and/or modify it at your own risk.                   *)
(* Version: 1.0                                                               *)
(*----------------------------------------------------------------------------*)
unit langsup;

interface
uses
   classes,
   jphash;

type
   TLangSup = class
   protected
      hash     : TJPHashTable;
      inverse  : TJPHashTable;
      FFilename: string;
      FLanguage: string;
      FContext : string;
      FNativeLanguage: string;
      FReadOnly: boolean;
      procedure split(s: string; var language,context,text,translated: string);
   public
      constructor Create(NativeLanguage: string);
      destructor  Destroy; override;
      procedure   LoadFromFile(filename: string); virtual;
      procedure   SaveToFile(filename: string); virtual;
      procedure   SetLanguage(Language: string);
      function    GetLanguage: string;
      function    GetContext: string;
      procedure   SetContext(Context: string);
      procedure   SetNativeLanguage(Language: string);
      function    Get(text: string): string;
      procedure   Put(native,translated: string);
      procedure   Delete(native: string);
      procedure   SetCaptions(Component: TComponent); virtual;
      function    GetFilename: string;
      procedure   GetLanguages(lis: TStringList);
      procedure   GetContextAndText(lis: TStringList);
      procedure   SetReadOnly(value: boolean);
   end;

implementation
uses
   stdctrls, menus, forms, buttons, tabnotbk, extctrls, sysutils;

constructor TLangSup.Create(NativeLanguage: string);
begin
  inherited Create;
  hash      := TJPHashTable.Create;
  inverse   := TJPHashTable.Create;
  FFilename := '';
  FReadOnly := false;
  FNativeLanguage := NativeLanguage;
end;


destructor TLangSup.Destroy;
begin
  inverse.Free;
  hash.Free;
  inherited Destroy;
end;


procedure TLangSup.split(s: string; var language,context,text,translated: string);
var i,part: integer;
begin
  language   := '';
  context    := '';
  text       := '';
  translated := '';
  part := 0;
  for i := 1 to length(s) do
      if (s[i] = '|') then
         inc(part)
      else
      if (part = 0) then
         language := language + s[i]
      else
      if (part = 1) then
         context := context + s[i]
      else
      if (part = 2) then
         text := text + s[i]
      else
      if (part = 3) then
         translated := translated + s[i];
end;


(*----------------------------------------------------------------------------*)
(* You might override this function and load the translations from a database *)
(*----------------------------------------------------------------------------*)
procedure TLangSup.LoadFromFile(filename: string);
var sl: TStringList;
    language,context,text,translated: string;
    i: integer;
begin
  sl := TStringList.Create;
  try
     FFilename := filename;
     sl.LoadFromFile(filename);
     for i := 0 to sl.Count - 1 do
         begin
           split(sl[i],language,context,text,translated);

           hash    . Add(language        + '|' + context + '|' + text, translated);
           inverse . Add(FNativeLanguage + '|' + context + '|' + translated, text);
         end;
  finally
     sl.Free;
  end;
end;

procedure TLangSup.SaveToFile(filename: string);
var sl: TStringList;
    i : integer;
begin
  sl := TStringList.Create;
  try
    hash.Keys(sl);
    for i := 0 to sl.Count - 1 do
        sl[i] := sl[i] + '|' + hash.Get(sl[i]);
    sl.Sorted := true;
    sl.SaveToFile(filename);
  finally
    sl.Free;
  end;
end;


procedure TLangSup.SetLanguage(Language: string);
begin
  FLanguage := Language;
end;

function TLangSup.GetLanguage: string;
begin
  result := FLanguage;
end;

procedure TLangSup.SetContext(Context: string);
begin
  FContext := Context;
end;

function TLangSup.GetContext: string;
begin
  result := FContext;
end;

procedure TLangSup.SetNativeLanguage(Language: string);
begin
  FNativeLanguage := Language;
end;








(*----------------------------------------------------------------------------*)
(* Get the translation of the string s. Language and context are given        *)
(* implicitly by FLanguage and FContext.                                      *)
(*----------------------------------------------------------------------------*)
function TLangSup.get(text: string): string;
var t: string;
begin
  if (FLanguage = '') then
     begin
       result := text;
       exit;
     end;

  if (text = '') then
     begin
       result := '';
       exit;
     end;

  t := FLanguage + '|' + FContext + '|' + text;
  if (hash.Find(t,result)) then
     exit; // Found the translation

  // Second try. text might already be in a foreign language.
  t := FLanguage + '|' + FContext + '|' + text;
  if (not inverse.Find(t,result)) then
     begin
       result := '???';
       Put(text,'<<<'+text+'>>>');
     end;

  if (FLanguage = FNativeLanguage) then
     exit;

  // Now result is in native language. We must translate it again.
  t := FLanguage + '|' + FContext + '|' + result;
  if (not hash.Find(t,result)) then
     begin
       result := '???';
       Put(text,'<<<'+text+'>>>');
     end;
end;


(*----------------------------------------------------------------------------*)
(* Store a pair text / text translated.                                       *)
(* Language and context are given implicitly.                                 *)
(*----------------------------------------------------------------------------*)
procedure TLangSup.Put(native,translated: string);
begin
  if (FReadOnly) then
     exit;

  if ((native = '') or (translated = '')) then
     exit;

  if (FLanguage = '') then
     exit;

  if (FContext = '') then
     exit;

  hash.Add(FLanguage + '|' + FContext + '|' + native, translated);
end;


(*----------------------------------------------------------------------------*)
(* Delete a text; language and context are given implicitly.                  *)
(*----------------------------------------------------------------------------*)
procedure TLangSup.Delete(native: string);
begin
  hash.Delete(FLanguage + '|' + FContext + '|' + native);
end;


(*----------------------------------------------------------------------------*)
(* Walk through a component and set all captions.                             *)
(* In the case something is missing, simply put in the code here or derive    *)
(* the class and override this procedure; it's virtual.                       *)
(*----------------------------------------------------------------------------*)
procedure TLangSup.SetCaptions(Component: TComponent);
var i: integer;
    s: string;
begin
  if (Component is TForm) then
     begin
       TForm(Component).Caption := Get(TForm(Component).Caption);
       for i := 0 to TForm(Component).ComponentCount - 1 do
           SetCaptions(TForm(Component).Components[i]);
     end
  else
  if (Component is TLabel) then
     TLabel(Component).Caption := Get(TLabel(Component).Caption)
  else
  if (Component is TButton) then
     TButton(Component).Caption := Get(TButton(Component).Caption)
  else
  if (Component is TBitBtn) then
     TBitBtn(Component).Caption := Get(TBitBtn(Component).Caption)
  else
  if (Component is TPanel) then
     TPanel(Component).Caption := Get(TPanel(Component).Caption)
  else
  if (Component is TCheckBox) then
     TCheckBox(Component).Caption := Get(TCheckBox(Component).Caption)
  else
  if (Component is TRadioGroup) then
     begin
       TRadioGroup(Component).Caption := Get(TRadioGroup(Component).Caption);
       for i := 0 to TRadioGroup(Component).Items.Count - 1 do
           TRadioGroup(Component).Items[i] := Get(TRadioGroup(Component).Items[i]);
     end
  else
  if (Component is TGroupBox) then
     TGroupBox(Component).Caption := Get(TGroupBox(Component).Caption)
  else
  if (Component is TMainMenu) then
     begin
       for i := 0 to TMainMenu(Component).ComponentCount - 1 do
           SetCaptions(TMainMenu(Component).Components[i]);
     end
  else
  if (Component is TTabbedNotebook) then
     begin
       for i := 0 to TTabbedNotebook(Component).Pages.Count - 1 do
           begin
             s := Get(TTabbedNotebook(Component).Pages[i]);
             TTabbedNotebook(Component).Pages[i] := s;
           end
     end
  else
  if (Component is TMenuItem) then
     begin
       TMenuItem(Component).Caption := Get(TMenuItem(Component).Caption);
       for i := 0 to TMenuItem(Component).ComponentCount - 1 do
           SetCaptions(TMenuItem(Component).Components[i]);
     end;
end;

function TLangSup.GetFilename: string;
begin
  result := FFilename;
end;


(*----------------------------------------------------------------------------*)
(* Get all known languages.                                                   *)
(*----------------------------------------------------------------------------*)
procedure TLangSup.GetLanguages(lis: TStringList);
var i : integer;
    sl: TStringList;
begin
  sl := TStringList.Create;
  try
    lis.Sorted     := true;
    lis.Duplicates := dupIgnore;
    lis.Add(FNativeLanguage);

    hash.Keys(sl);
    for i := 0 to sl.Count - 1 do
        if (pos('|',sl[i]) <> 0) then
           lis.Add(copy(sl[i],1,pos('|',sl[i])-1));
  finally
    sl.Free;
  end;
end;

(*----------------------------------------------------------------------------*)
(* Get a list of all combinations context+word, regardless of the language.   *)
(*----------------------------------------------------------------------------*)
procedure TLangSup.GetContextAndText(lis: TStringList);
var s,toAdd: string;
    i,p    : integer;
    sl     : TStringList;
begin
  sl := TStringList.Create;
  try
    hash.Keys(sl);
    lis.Sorted     := true;
    lis.Duplicates := dupIgnore;
    for i := 0 to sl.Count - 1 do
        begin
          s := sl.Strings[i];
          p := pos('|',s);
          toAdd := copy(s,p+1,length(s));
          lis.Add(toAdd);
        end;
  finally
    sl.Free;
  end;
end;


procedure TLangSup.SetReadOnly(value: boolean);
begin
  FReadOnly := value;
end;



end.
